
 
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>

#include <GL/glew.h> 

#include <GL/glut.h> 
#include "SDL.h"

#define GL_GLEXT_PROTOTYPES 1

/* screen width, height, and bit depth */
#define SCREEN_WIDTH  1024
#define SCREEN_HEIGHT 768
#define SCREEN_BPP     16

/* Set up some booleans */
#define TRUE  1
#define FALSE 0

/* This is our SDL surface */
SDL_Surface *surface;

/* Whether or not lighting is on */
int light = FALSE;

GLfloat xrot = 0;      /* X Rotation */
GLfloat yrot = 0;      /* Y Rotation */
GLfloat xspeed = 1.0f;    /* X Rotation Speed */
GLfloat yspeed = 1.0f;    /* Y Rotation Speed */
GLfloat z = -8.0f; /* Depth Into The Screen */


/* Ambient Light Values ( NEW ) */
GLfloat LightAmbient[]  = { 0.5f, 0.5f, 0.5f, 1.0f };
/* Diffuse Light Values ( NEW ) */
GLfloat LightDiffuse[]  = { 1.0f, 1.0f, 1.0f, 1.0f };
/* Light Position ( NEW ) */
GLfloat LightPosition[] = { 0.0f, 0.0f, 2.0f, 1.0f };

GLuint filter = 0;     /* Which Filter To Use */
GLuint texture[3]; /* Storage for 3 textures */

GLuint program[3];

GLuint vertexShader;
GLuint fragmentShader;

GLint xArgument;
GLint yArgument;
GLint timeArgument;

int program_cur = 0;
int program_max = 0;

int shader_active = FALSE;
int blend_active = FALSE;
int mouse_x = 0;
int mouse_y = 0;
float mouse_relx = 0.0f;
float mouse_rely = 0.0f;
int mouse_left_down = FALSE;
int mouse_right_down = FALSE;

void checkError(GLuint handle) {
    GLsizei infologLength = 0;
    GLsizei charsWritten  = 0;
    char *infoLog;

    glGetShaderiv(handle, GL_INFO_LOG_LENGTH,
                              &infologLength);

    if (infologLength > 1)
    {
        infoLog = (char*)malloc(infologLength);
        glGetShaderInfoLog(handle, infologLength, &charsWritten, infoLog);
        fprintf(stderr, "Error: %s", infoLog);
        free(infoLog);
    }
}
      

char * loadShaderSource(const char *fn) {
   
  int length;
  char *buf;
  SDL_RWops *rw=SDL_RWFromFile(fn,"rb");
  if(rw==NULL)
  {
    fprintf(stderr,"Couldn't open %s\n",fn);
    return NULL;
  }
  length=SDL_RWseek(rw,0,SEEK_END);
  if(length<0)
  {
    fprintf(stderr,"Could not seek inside %s\n",fn);
    return NULL;
  }
  fprintf(stderr,"%s is %d bytes long\n",fn,length);
  SDL_RWseek(rw,0,SEEK_SET);
  buf = (char*)malloc(length * sizeof(char) + 1);
  SDL_RWread(rw,buf,1,length);
  buf[length] = 0;
  SDL_RWclose(rw);
  return buf;
}      
      

/* function to release/destroy our resources and restoring the old desktop */
void Quit( int returnCode )
{
    /* clean up the window */
    SDL_Quit( );

    /* and exit appropriately */
    exit( returnCode );
}

/* function to load in bitmap as a GL texture */
int LoadGLTextures( )
{
    /* Status indicator */
    int Status = FALSE;

    /* Create storage space for the texture */
    SDL_Surface *TextureImage[1]; 

    /* Load The Bitmap, Check For Errors, If Bitmap's Not Found Quit */
    if ( ( TextureImage[0] = SDL_LoadBMP( "data/glass.bmp" ) ) )
        {

	    /* Set the status to true */
	    Status = TRUE;

	    /* Create The Texture */
	    glGenTextures( 3, &texture[0] );

	    /* Load in texture 1 */
	    /* Typical Texture Generation Using Data From The Bitmap */
	    glBindTexture( GL_TEXTURE_2D, texture[0] );

	    /* Generate The Texture */
	    glTexImage2D( GL_TEXTURE_2D, 0, 3, TextureImage[0]->w,
			  TextureImage[0]->h, 0, GL_BGR,
			  GL_UNSIGNED_BYTE, TextureImage[0]->pixels );
	    
	    /* Nearest Filtering */
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			     GL_NEAREST );
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
			     GL_NEAREST );

	    /* Load in texture 2 */
	    /* Typical Texture Generation Using Data From The Bitmap */
	    glBindTexture( GL_TEXTURE_2D, texture[1] );

	    /* Linear Filtering */
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			     GL_LINEAR );
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
			     GL_LINEAR );

	    /* Generate The Texture */
	    glTexImage2D( GL_TEXTURE_2D, 0, 3, TextureImage[0]->w,
			  TextureImage[0]->h, 0, GL_BGR,
			  GL_UNSIGNED_BYTE, TextureImage[0]->pixels );

	    /* Load in texture 3 */
	    /* Typical Texture Generation Using Data From The Bitmap */
	    glBindTexture( GL_TEXTURE_2D, texture[2] );

	    /* Mipmapped Filtering */
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			     GL_LINEAR_MIPMAP_NEAREST );
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
			     GL_LINEAR );

	    /* Generate The MipMapped Texture ( NEW ) */
	    gluBuild2DMipmaps( GL_TEXTURE_2D, 3, TextureImage[0]->w,
			       TextureImage[0]->h, GL_BGR,
			       GL_UNSIGNED_BYTE, TextureImage[0]->pixels );
        }

    /* Free up any memory we may have used */
    if ( TextureImage[0] )
	    SDL_FreeSurface( TextureImage[0] );

    return Status;
}

/* function to reset our viewport after a window resize */
int resizeWindow( int width, int height )
{
    /* Height / width ration */
    GLfloat ratio;
 
    /* Protect against a divide by zero */
    if ( height == 0 )
	height = 1;

    ratio = ( GLfloat )width / ( GLfloat )height;

    /* Setup our viewport. */
    glViewport( 0, 0, ( GLint )width, ( GLint )height );

    /* change to the projection matrix and set our viewing volume. */
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity( );

    /* Set our perspective */
    gluPerspective( 45.0f, ratio, 0.1f, 100.0f );

    /* Make sure we're chaning the model view and not the projection */
    glMatrixMode( GL_MODELVIEW );

    /* Reset The View */
    glLoadIdentity( );

    return( TRUE );
}

/* function to handle key press events */
void handleKeyPress( SDL_keysym *keysym )
{
    switch ( keysym->sym )
	{
	case SDLK_ESCAPE:
	    /* ESC key was pressed */
	    Quit( 0 );
	    break;
	case SDLK_f:
	    /* 'f' key was pressed
	     * this pages through the different filters
	     */
	    filter = ( ++filter ) % 3;
	    break;
    
    case SDLK_n:
        /* 'f' key was pressed
         * this pages through the different filters
         */
        program_cur = ( ++program_cur ) % (program_max);
        shader_active = TRUE;
        printf("programm %i\n",program_cur);
        glUseProgram(program[program_cur]);
        break;

  
    case SDLK_l:
	    /* 'l' key was pressed
	     * this toggles the light
	     */
	    light = !light;
	    if ( !light )
		glDisable( GL_LIGHTING );
	    else
		glEnable( GL_LIGHTING );
	    break;
    case SDLK_s:
        /* 's' key was pressed
         * toggle shader programm
         */
        if (shader_active){
            shader_active = FALSE;
            glUseProgram(0);
        } else {
            shader_active = TRUE;
            glUseProgram(program[program_cur]);
        }
        break;
	case SDLK_PAGEUP:
	    /* PageUp key was pressed
	     * this zooms into the scene
	     */
	    z -= 0.02f;
	    break;
	case SDLK_PAGEDOWN:
	    /* PageDown key was pressed
	     * this zooms out of the scene
	     */
	    z += 0.02f;
	    break;
	case SDLK_UP:
	    /* Up arrow key was pressed
	     * this affects the x rotation
	     */
	    xspeed -= 0.01f;
	    break;
	case SDLK_DOWN:
	    /* Down arrow key was pressed
	     * this affects the x rotation
	     */
	    xspeed += 0.01f;
	    break;
	case SDLK_RIGHT:
	    /* Right arrow key was pressed
	     * this affects the y rotation
	     */
	    yspeed += 0.01f;
	    break;
	case SDLK_LEFT:
	    /* Left arrow key was pressed
	     * this affects the y rotation
	     */
	    yspeed -= 0.01f;
	    break;
	case SDLK_F1:
	    /* 'f' key was pressed
	     * this toggles fullscreen mode
	     */
	    SDL_WM_ToggleFullScreen( surface );
	    break;
    case SDLK_b:
        if (blend_active){
            blend_active = FALSE;
            glDisable(GL_BLEND);            /* Turn Blending Off */
            glEnable(GL_DEPTH_TEST);        /* Turn Depth Testing On */
        } else {
            glEnable(GL_BLEND);             /* Turn Blending On */
            glDisable(GL_DEPTH_TEST);       /* Turn Depth Testing Off */
            blend_active = TRUE;
            
        }
        break;
    default:
	    break;
	}

    return;
}

void initShader() {
    char *vs,*fs;
    int i;
    GLenum err;
    shader_active = FALSE;
    err = glewInit();
    char * names[5] = {"deform", "uniform", "toonshading", "gouraud", 0 };
    char name[256];
    
    if (GLEW_OK != err){
        /* Problem: glewInit failed, something is seriously wrong. */
        fprintf(stderr, "Error: %s\n", glewGetErrorString(err));
        return;
    }
    if (!GLEW_ARB_vertex_program)
    {
        fprintf(stderr, "Error: no GLEW_ARB_vertex_program\n");
    }
  
    shader_active = TRUE;
    
    for (i=0; names[i]; i++){  
        program[i] = glCreateProgram();
        vertexShader = glCreateShader(GL_VERTEX_SHADER);
        fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
        printf("Programm %i\n",i);
        sprintf(name,"data/%s.vert",names[i]);
        vs = loadShaderSource(name);
        sprintf(name,"data/%s.frag",names[i]);
        fs = loadShaderSource(name);
        
        const char* vv = vs;
        const char* ff = fs;
    
        glShaderSource(vertexShader, 1, &vv,NULL);
        checkError(vertexShader);
        free(vs);
    
        glShaderSource(fragmentShader, 1, &ff,NULL);
        checkError(fragmentShader);
        free(fs);
    
        glCompileShader(vertexShader);
        checkError(vertexShader);
    
        glCompileShader(fragmentShader);
        checkError(fragmentShader);
    
        glAttachShader(program[i],vertexShader);
        glAttachShader(program[i],fragmentShader);
    
        glLinkProgram(program[i]);
    
        checkError(program[i]);
        timeArgument = glGetUniformLocation(program[i], "time");
    } 
    program_max = i;
    glUseProgram(program[0]);
}


/* general OpenGL initialization function */
int initGL( GLvoid )
{

    /* Load in the texture */
    if ( !LoadGLTextures( ) )
	   return FALSE;
    /* Enable Texture Mapping ( NEW ) */
    glEnable( GL_TEXTURE_2D );

    /* Enable smooth shading */
    glShadeModel( GL_SMOOTH );

    /* Set the background black */
    glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );

    /* Depth buffer setup */
    glClearDepth( 1.0f );

    /* Enables Depth Testing */
    glEnable( GL_DEPTH_TEST );

    /* The Type Of Depth Test To Do */
    glDepthFunc( GL_LEQUAL );

    /* Really Nice Perspective Calculations */
    glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

    /* Setup The Ambient Light */
    glLightfv( GL_LIGHT1, GL_AMBIENT, LightAmbient );

    /* Setup The Diffuse Light */
    glLightfv( GL_LIGHT1, GL_DIFFUSE, LightDiffuse );

    /* Position The Light */
    glLightfv( GL_LIGHT1, GL_POSITION, LightPosition );

    /* Enable Light One */
    glEnable( GL_LIGHT1 );

    /* Full Brightness, 50% Alpha ( NEW ) */
    glColor4f( 1.0f, 1.0f, 1.0f, 0.5f);

    /* Blending Function For Translucency Based On Source Alpha Value  */
    glBlendFunc( GL_SRC_ALPHA, GL_ONE );

    return( TRUE );
}

/* Here goes our drawing code */
int drawGLScene( GLvoid )
{
    
    if (yspeed != 0.0f){
        yspeed -= yspeed/1000.0f;
    }
    if (xspeed != 0.0f){
        xspeed -= xspeed/1000.0f;
    }
      
    /* These are to calculate our fps */
    static GLint T0     = 0;
    static GLint Frames = 0;

    /* Clear The Screen And The Depth Buffer */
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    if (program_cur == 0){
        glUniform1f (xArgument, mouse_relx);
        glUniform1f (yArgument, mouse_rely);
    }
    if (program_cur == 1){
        glUniform1f (timeArgument, xrot);
    }
    /* Reset the view */
    glLoadIdentity( );

    /* Translate Into/Out Of The Screen By z */
    glTranslatef( +1.5f, 0.0f, z );

    glRotatef( xrot, 1.0f, 0.0f, 0.0f); /* Rotate On The X Axis By xrot */
    glRotatef( yrot, 0.0f, 1.0f, 0.0f); /* Rotate On The Y Axis By yrot */

    /* Select A Texture Based On filter */
    glBindTexture( GL_TEXTURE_2D, texture[filter] );

    /* Start Drawing Quads */
    glBegin( GL_QUADS );
      /* Front Face */
      /* Normal Pointing Towards Viewer */
      glNormal3f( 0.0f, 0.0f, 1.0f );
      /* Point 1 (Front) */
      glTexCoord2f( 1.0f, 0.0f ); glVertex3f( -1.0f, -1.0f,  1.0f );
      /* Point 2 (Front) */
      glTexCoord2f( 0.0f, 0.0f ); glVertex3f(  1.0f, -1.0f,  1.0f );
      /* Point 3 (Front) */
      glTexCoord2f( 0.0f, 1.0f ); glVertex3f(  1.0f,  1.0f,  1.0f );
      /* Point 4 (Front) */
      glTexCoord2f( 1.0f, 1.0f ); glVertex3f( -1.0f,  1.0f,  1.0f );
      
      /* Back Face */
      /* Normal Pointing Away From Viewer */
      glNormal3f( 0.0f, 0.0f, -1.0f);
      /* Point 1 (Back) */
      glTexCoord2f( 0.0f, 0.0f ); glVertex3f( -1.0f, -1.0f, -1.0f );
      /* Point 2 (Back) */
      glTexCoord2f( 0.0f, 1.0f ); glVertex3f( -1.0f,  1.0f, -1.0f );
      /* Point 3 (Back) */
      glTexCoord2f( 1.0f, 1.0f ); glVertex3f(  1.0f,  1.0f, -1.0f );
      /* Point 4 (Back) */
      glTexCoord2f( 1.0f, 0.0f ); glVertex3f(  1.0f, -1.0f, -1.0f );

      /* Top Face */
      /* Normal Pointing Up */
      glNormal3f( 0.0f, 1.0f, 0.0f );
      /* Point 1 (Top) */
      glTexCoord2f( 1.0f, 1.0f ); glVertex3f( -1.0f,  1.0f, -1.0f );
      /* Point 2 (Top) */
      glTexCoord2f( 1.0f, 0.0f ); glVertex3f( -1.0f,  1.0f,  1.0f );
      /* Point 3 (Top) */
      glTexCoord2f( 0.0f, 0.0f ); glVertex3f(  1.0f,  1.0f,  1.0f );
      /* Point 4 (Top) */
      glTexCoord2f( 0.0f, 1.0f ); glVertex3f(  1.0f,  1.0f, -1.0f );

      /* Bottom Face */
      /* Normal Pointing Down */
      glNormal3f( 0.0f, -1.0f, 0.0f );
      /* Point 1 (Bottom) */
      glTexCoord2f( 0.0f, 1.0f ); glVertex3f( -1.0f, -1.0f, -1.0f );
      /* Point 2 (Bottom) */
      glTexCoord2f( 1.0f, 1.0f ); glVertex3f(  1.0f, -1.0f, -1.0f );
      /* Point 3 (Bottom) */
      glTexCoord2f( 1.0f, 0.0f ); glVertex3f(  1.0f, -1.0f,  1.0f );
      /* Point 4 (Bottom) */
      glTexCoord2f( 0.0f, 0.0f ); glVertex3f( -1.0f, -1.0f,  1.0f );

      /* Right face */
      /* Normal Pointing Right */
      glNormal3f( 1.0f, 0.0f, 0.0f);
      /* Point 1 (Right) */
      glTexCoord2f( 0.0f, 0.0f ); glVertex3f( 1.0f, -1.0f, -1.0f );
      /* Point 2 (Right) */
      glTexCoord2f( 0.0f, 1.0f ); glVertex3f( 1.0f,  1.0f, -1.0f );
      /* Point 3 (Right) */
      glTexCoord2f( 1.0f, 1.0f ); glVertex3f( 1.0f,  1.0f,  1.0f );
      /* Point 4 (Right) */
      glTexCoord2f( 1.0f, 0.0f ); glVertex3f( 1.0f, -1.0f,  1.0f );

      /* Left Face*/
      /* Normal Pointing Left */
      glNormal3f( -1.0f, 0.0f, 0.0f );
      /* Point 1 (Left) */
      glTexCoord2f( 1.0f, 0.0f ); glVertex3f( -1.0f, -1.0f, -1.0f );
      /* Point 2 (Left) */
      glTexCoord2f( 0.0f, 0.0f ); glVertex3f( -1.0f, -1.0f,  1.0f );
      /* Point 3 (Left) */
      glTexCoord2f( 0.0f, 1.0f ); glVertex3f( -1.0f,  1.0f,  1.0f );
      /* Point 4 (Left) */
      glTexCoord2f( 1.0f, 1.0f ); glVertex3f( -1.0f,  1.0f, -1.0f );

    glEnd();
   
    /* Reset the view */
    glLoadIdentity( );

    glTranslatef( -1.5f, 0.0f, z );

    glRotatef( xrot, 1.0f, 0.0f, 0.0f); /* Rotate On The X Axis By xrot */
    glRotatef( yrot, 0.0f, 1.0f, 0.0f); /* Rotate On The Y Axis By yrot */

    glutSolidTeapot(1);

    /* Draw it to the screen */
    SDL_GL_SwapBuffers( );

    /* Gather our frames per second */
    Frames++;
    {
	GLint t = SDL_GetTicks();
	if (t - T0 >= 5000) {
	    GLfloat seconds = (t - T0) / 1000.0;
	    GLfloat fps = Frames / seconds;
	    printf("%d frames in %g seconds = %g FPS\n", Frames, seconds, fps);
	    T0 = t;
	    Frames = 0;
	}
    }

    xrot += xspeed; /* Add xspeed To xrot */
    yrot += yspeed; /* Add yspeed To yrot */

    return( TRUE );
}

int main( int argc, char **argv )
{

    /* Flags to pass to SDL_SetVideoMode */
    int videoFlags;
    /* main loop variable */
    int done = FALSE;
    /* used to collect events */
    SDL_Event event;
    /* this holds some info about our display */
    const SDL_VideoInfo *videoInfo;
    /* whether or not the window is active */
    int isActive = TRUE;

    /* initialize SDL */
    if ( SDL_Init( SDL_INIT_VIDEO ) < 0 )
	{
	    fprintf( stderr, "Video initialization failed: %s\n",
		     SDL_GetError( ) );
	    Quit( 1 );
	}

    /* Fetch the video info */
    videoInfo = SDL_GetVideoInfo( );

    if ( !videoInfo )
	{
	    fprintf( stderr, "Video query failed: %s\n",
		     SDL_GetError( ) );
	    Quit( 1 );
	}

    /* the flags to pass to SDL_SetVideoMode */
    videoFlags  = SDL_OPENGL;          /* Enable OpenGL in SDL */
    videoFlags |= SDL_GL_DOUBLEBUFFER; /* Enable double buffering */
    videoFlags |= SDL_HWPALETTE;       /* Store the palette in hardware */
    videoFlags |= SDL_RESIZABLE;       /* Enable window resizing */

    /* This checks to see if surfaces can be stored in memory */
    if ( videoInfo->hw_available )
	videoFlags |= SDL_HWSURFACE;
    else
	videoFlags |= SDL_SWSURFACE;

    /* This checks if hardware blits can be done */
    if ( videoInfo->blit_hw )
	videoFlags |= SDL_HWACCEL;

    /* Sets up OpenGL double buffering */
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

    /* get a SDL surface */
    surface = SDL_SetVideoMode( SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP,
				videoFlags );

    /* Verify there is a surface */
    if ( !surface )
	{
	    fprintf( stderr,  "Video mode set failed: %s\n", SDL_GetError( ) );
	    Quit( 1 );
	}
    /* initialize OpenGL */
    initGL( );
    initShader();
    
    /* resize the initial window */
    resizeWindow( SCREEN_WIDTH, SCREEN_HEIGHT );

    /* wait for events */
    while ( !done )
	{
	    /* handle the events in the queue */

	    while ( SDL_PollEvent( &event ) )
		{
		    switch( event.type )
			{
			case SDL_ACTIVEEVENT:
			    /* Something's happend with our focus
			     * If we lost focus or we are iconified, we
			     * shouldn't draw the screen
			     */
			    if ( event.active.gain == 0 )
				isActive = FALSE;
			    else
				isActive = TRUE;
			    break;			    
			case SDL_VIDEORESIZE:
			    /* handle resize event */
			    surface = SDL_SetVideoMode( event.resize.w,
							event.resize.h,
							16, videoFlags );
			    if ( !surface )
				{
				    fprintf( stderr, "Could not get a surface after resize: %s\n", SDL_GetError( ) );
				    Quit( 1 );
				}
			    resizeWindow( event.resize.w, event.resize.h );
			    break;
			case SDL_KEYDOWN:
			    /* handle key presses */
			    handleKeyPress( &event.key.keysym );
			    break;
			case SDL_QUIT:
			    /* handle quit requests */
			    done = 1;
			    break;
            case SDL_MOUSEMOTION:
                if (mouse_left_down){
                    if (event.button.x > mouse_x){
                        yspeed -= 0.01f;
                    } else if (event.button.x < mouse_x){
                        yspeed += 0.01f;
                    }
                    if (event.button.y > mouse_y){
                        xspeed += 0.01f;
                    } else if (event.button.y < mouse_y){
                        xspeed -= 0.01f;
                    }
                 mouse_relx = (mouse_x - SCREEN_WIDTH / 2.0f) / SCREEN_WIDTH;
                 mouse_rely = -(mouse_y - SCREEN_HEIGHT / 2.0f) / SCREEN_HEIGHT;   
                }
                if (mouse_right_down){
                    if (event.button.y > mouse_y){
                        z += 0.1f;
                    } else if (event.button.y < mouse_y){
                        z -= 0.1f;
                    }
                 }
                 mouse_x = event.button.x;
                 mouse_y = event.button.y;
                 break;
            case SDL_MOUSEBUTTONDOWN:
                if (event.button.button == SDL_BUTTON_LEFT){
                    mouse_left_down = TRUE;
                } else if (event.button.button == SDL_BUTTON_RIGHT){
                    mouse_right_down = TRUE;
                }
                
                break;
            case SDL_MOUSEBUTTONUP:
                if (event.button.button == SDL_BUTTON_LEFT){
                    mouse_left_down = FALSE;
                }else if (event.button.button == SDL_BUTTON_RIGHT){
                    mouse_right_down = FALSE;
                }
                break;
			default:
			    break;
			}
		}
	    /* draw the scene */
	    if ( isActive )
		drawGLScene( );
	}
    /* clean ourselves up and exit */
    Quit( 0 );

    /* Should never get here */
    return( 0 );
}
